package guess

import (
	"errors"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/aac"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/flac"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/mp3"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/opus"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/tta"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/vorbis"
	"io"
	"path"
	"strings"
)

var knownFormats = []format.Format{
	aac.NewFormat(),
	flac.NewFormat(),
	mp3.NewFormat(),
	opus.NewFormat(),
	tta.NewFormat(),
	vorbis.NewFormat(),
}

func GetDecoders(r io.ReadSeekCloser, pathName string) (decoders []format.Decoder, err error) {
	extension := strings.ToLower(strings.TrimPrefix(path.Ext(pathName), "."))

	var startOffset int64
	startOffset, err = r.Seek(0, io.SeekCurrent)
	if err != nil {
		return
	}
	peek := make([]byte, 16)
	if _, err = r.Read(peek); err != nil {
		return
	}
	if _, err = r.Seek(startOffset, io.SeekStart); err != nil {
		return
	}

	for _, formatEntry := range knownFormats {
		if decoder, ok := formatEntry.(format.Decoder); ok {
			if decoder.Identify(peek, extension) {
				decoders = append(decoders, decoder)
			}
		}
	}

	if len(decoders) == 0 {
		return nil, errors.New("could not find decoder")
	}

	return
}

func Open(r io.ReadSeekCloser, decoders []format.Decoder) (source audio.Source, err error) {
	var startOffset int64
	startOffset, err = r.Seek(0, io.SeekCurrent)

	for _, decoder := range decoders {
		if _, err = r.Seek(startOffset, io.SeekStart); err != nil {
			return
		}
		source, err = decoder.Open(r)
		if source.Blocks != nil && err == nil {
			return
		}
	}

	//seek back
	if _, err = r.Seek(startOffset, io.SeekStart); err != nil {
		return
	}
	return audio.Source{}, errors.New("could not open stream")
}

func OpenAnalyzer(r io.ReadSeekCloser, decoders []format.Decoder) (source audio.Source, analyzerChannel format.AnalyzerChannel, err error) {
	var startOffset int64
	startOffset, err = r.Seek(0, io.SeekCurrent)

	for _, decoder := range decoders {
		if analyzer, ok := decoder.(format.AnalyzerDecoder); ok {
			if _, err = r.Seek(startOffset, io.SeekStart); err != nil {
				return
			}
			source, analyzerChannel, err = analyzer.OpenAnalyzer(r)
			if source.Blocks != nil && err == nil {
				return
			}
		}
	}

	//seek back
	if _, err = r.Seek(startOffset, io.SeekStart); err != nil {
		return
	}
	return audio.Source{}, nil, errors.New("could not open stream analyzer")
}
